/*
 * Chapter1Figs.cpp - Version 1.1 - March, 2019
 *
 * Example program illustrating many of the figures and
 *    sample C++ and GLSL code in Chapter 1 of the book below.
 *
 * Author: Sam Buss
 *
 * Software accompanying POSSIBLE SECOND EDITION TO the book
 *		3D Computer Graphics: A Mathematical Introduction with OpenGL,
 *		by S. Buss, Cambridge University Press, 2003.
 *
 * Software is "as-is" and carries no warranty.  It may be used without
 *   restriction, but if you modify it, please change the filenames to
 *   prevent confusion between different versions.
 * Bug reports: Sam Buss, sbuss@ucsd.edu.
 * Web page: http://math.ucsd.edu/~sbuss/MathCG2
 */

// Use space to toggle what image is shown.
// Use Escape or 'X' or 'x' to exit.

// These libraries are needed to link the program (Visual Studio specific)
#pragma comment(lib,"opengl32.lib")
#pragma comment(lib,"glu32.lib")
#pragma comment(lib,"glfw3.lib")
#pragma comment(lib,"glew32s.lib")
#pragma comment(lib,"glew32.lib")

// Use the static library (so glew32.dll is not needed):
#define GLEW_STATIC
#include <GL/glew.h> 
#include <GLFW/glfw3.h>

#include "GlShaderMgr.h"
bool check_for_opengl_errors();     // Function prototype (should really go in a header file)

// Enable standard input and output via printf(), etc.
// Put this include *after* the includes for glew and GLFW!
#include <stdio.h>

// ********************
// Animation controls and state infornation
// ********************

int CurrentMode = 0;	// Controls what is drawn.

// ************************
// General data helping with setting up VAO (Vertex Array Objects)
//    and Vertex Buffer Objects.
// ***********************

const int NumObjects = 3;
const int sixPoints = 0;
const int sevenPoints = 1;
const int colorTri = 2;

unsigned int myVBO[NumObjects];  // Vertex Buffer Object - holds an array of data
unsigned int myVAO[NumObjects];  // Vertex Array Object - holds info about an array of vertex data;

// We create three shader programs
unsigned int shaderProgram1;   // Simplest shader, just passes through the color
unsigned int shaderProgram2;   // "Flat" colors only
unsigned int shaderProgram3;   // All colors set to maximum brightness
const unsigned int vertPos_loc = 0;   // Corresponds to "location = 0" in the verter shader definitions
const unsigned int vertColor_loc = 1; // Corresponds to "location = 1" in the verter shader definitions

// *************************
// mySetupGeometries defines the scene data, especially vertex  positions and colors.
//    - It also loads all the data into the VAO's (Vertex Array Objects) and
//      into the VBO's (Vertex Buffer Objects).
// This routine is only called once to initialize the data.
// *************************
void mySetupGeometries() {

	// In this simple example, we do not use the Projection or
	//   ModelView matrices. But the the vertex shader
    //   does an ad hoc rescaling.  Hence, all x, y, z positions
	//   in the arrays should be in the range [0,2]x[0,3]x[-1,1].

    // Allocate Vertex Array Objects (VAOs) and Vertex Buffer Objects (VBOs).
    glGenVertexArrays(NumObjects, &myVAO[0]);
    glGenBuffers(NumObjects, &myVBO[0]);

    // First geometry
    // Array of x,y coordinates for six vertices
    float verts[] = {
        0.5f, 1.0f, // Vertex v_0
        2.0f, 2.0f, // Vertex v_1
        1.8f, 2.6f, // Vertex v_2
        0.7f, 2.2f, // Vertex v_3
        1.6f, 1.2f, // Vertex v_4
        1.0f, 0.5f, // Vertex v_5
    };

    // Bind and load vertex data into vertex buffer.  
    //   This is stored in the GPU for quick access.
    glBindBuffer(GL_ARRAY_BUFFER, myVBO[sixPoints]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(verts), verts, GL_STATIC_DRAW);

    // Bind the Vertex Array Object and then configure vertex attributes(s).
    //   The vertices consist of three coordinates (x, y, z values)
	//     The x,y  values come from the  array. The z values default to 0.0.
    //   This information is stored in the VertexArrayObject
    //   vertPos_loc corresponds to the "location = 0" in the vertex shader.
    //          These positions are loaded from an array, one per vertex.
    //  The color values are not set here: this is done in myRenderScene().
    glBindVertexArray(myVAO[sixPoints]);
    glVertexAttribPointer(vertPos_loc, 2, GL_FLOAT, GL_FALSE, 0, (void*)0);	// Info about where positions are in the VBO
    glEnableVertexAttribArray(vertPos_loc);									// Enable the stored vertex positions

                                                                            // Array of x,y,z coordinates for seven vertices
    float verts2[] = {
        0.25, 0.5, 0.0, // Vertex w 0
        1.25f, 1.0f, 0.0, // Vertex w 1
        0.75f, 1.5f, 0.0, // Vertex w 2
        1.75f, 1.8f, 0.0, // Vertex w 3
        2.0f, 3.0f, 0.0, // Vertex w 5
        1.05f, 2.5f, 0.0, // Vertex w 4
        0.4f, 2.4f, 0.0, // Vertex w 6
    };
    unsigned int elements[] = {
        0, 1, 2, 3, 4, 5, // For GL TRIANGLES
        2, 0, 1, 3, 5, 6, // For GL TRIANGLE FAN
        0, 1, 2, 3, 5, 4 // For GL TRIANGLE STRIP
    };

    glBindVertexArray(myVAO[sevenPoints]);
    glBindBuffer(GL_ARRAY_BUFFER, myVBO[sevenPoints]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(verts2), verts2, GL_STATIC_DRAW);
    glBindVertexArray(myVAO[sevenPoints]);
    glVertexAttribPointer(vertPos_loc, 3, GL_FLOAT, GL_FALSE, 3*sizeof(float), (void*)0);
    glEnableVertexAttribArray(vertPos_loc);

    unsigned int theEBO;
    glGenBuffers(1, &theEBO);
    glBindVertexArray(myVAO[sevenPoints]);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, theEBO);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(elements), elements, GL_STATIC_DRAW);

    // Array of x,y coordinates and r,g,b colors for three vertices
    float verts3[] = {
        // x, y position; R, G, B color
        0.1f, 1.0,         1.0, 0.0, 0.0, // 1st vertex, Red 1,0,0
        1.9f, 1.0,         0.0, 1.0, 0.0, // 2nd vertex, Green 0,1,0
        1.0f, 2.5,         0.0, 0.0, 1.0, // 3rd vertex, Blue 0,0,1
    };

    glBindBuffer(GL_ARRAY_BUFFER, myVBO[colorTri]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(verts3), verts3, GL_STATIC_DRAW);
    glBindVertexArray(myVAO[colorTri]);
    glVertexAttribPointer(vertPos_loc, 2, GL_FLOAT, GL_FALSE,
        5 * sizeof(float), (void*)0);
    glVertexAttribPointer(vertColor_loc, 3, GL_FLOAT, GL_FALSE,
        5 * sizeof(float), (void*)(2 * sizeof(float)));
    glEnableVertexAttribArray(vertPos_loc);
    glEnableVertexAttribArray(vertColor_loc);

	// This is optional, but allowed.  The VAO already knows which buffer (VBO) is holding its
	//     vertex data, so it is OK to unbind the VBO here.
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	// Also optional, but perhaps a good idea so that the last VAO is not used by accident.
	glBindVertexArray(0);

    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}
// *************************************
// Main routine for rendering the scene
// myRenderScene() is called every time the scene needs to be redrawn.
// mySetupGeometries() has already created the vertex and buffer objects, etc.
// setup_shaders() has already created the shader programs.
// *************************************
void myRenderScene() {
    static const float white[] = { 1.0f, 1.0f, 1.0f, 1.0f };

	// Clear the rendering window
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// This very simple shader program is used for all the items.
	glUseProgram(shaderProgram1);

	switch (CurrentMode) {
    case 0:
        // Draw six points
        glBindVertexArray(myVAO[sixPoints]);
        glVertexAttrib3f(vertColor_loc, 1.0f, 0.5f, 0.2f);		// An orange-red color (R, G, B values).
        glDrawArrays(GL_POINTS, 0, 6);
        break;
    case 1:
        // Draw line segments
        glBindVertexArray(myVAO[sixPoints]);
        glVertexAttrib3f(vertColor_loc, 1.0f, 0.5f, 0.2f);		// An orange-red color (R, G, B values).
        glDrawArrays(GL_LINES, 0, 6);
        break;
    case 2:
        // Draw line strip
        glBindVertexArray(myVAO[sixPoints]);
        glVertexAttrib3f(vertColor_loc, 1.0f, 0.5f, 0.2f);		// An orange-red color (R, G, B values).
        glDrawArrays(GL_LINE_STRIP, 0, 6);
        break;
    case 3:
        // Draw line loop
        glBindVertexArray(myVAO[sixPoints]);
        glVertexAttrib3f(vertColor_loc, 1.0f, 0.5f, 0.2f);		// An orange-red color (R, G, B values).
        glDrawArrays(GL_LINE_LOOP, 0, 6);
        break;
    case 4:
        glBindVertexArray(myVAO[sevenPoints]);
        glVertexAttrib3f(vertColor_loc, 0.7f, 0.7f, 0.7f);		// Light gray.
        glDrawArrays(GL_TRIANGLES, 0, 6);
        break;
    case 5:
        // Drawing GL_TRIANGLES with glDrawElements instead of glDrawArrays.
        //      Identical effect as in case 4.
        glBindVertexArray(myVAO[sevenPoints]);
        glVertexAttrib3f(vertColor_loc, 0.7f, 0.7f, 0.7f); // Light gray
        glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0);
        break;
    case 6:
        glPolygonMode(GL_FRONT, GL_LINE);
        glBindVertexArray(myVAO[sevenPoints]);
        glVertexAttrib3f(vertColor_loc, 0.7f, 0.7f, 0.7f); // Light gray
        glDrawElements(GL_TRIANGLE_FAN, 6, GL_UNSIGNED_INT, (void*)(6 * sizeof(unsigned int)));
        break;
    case 7:
        glBindVertexArray(myVAO[sevenPoints]);
        glVertexAttrib3f(vertColor_loc, 0.7f, 0.7f, 0.7f); // Light gray
        glDrawElements(GL_TRIANGLE_STRIP, 6, GL_UNSIGNED_INT, (void*)(12 * sizeof(unsigned int)));
        break;
    case 8:
        // Draw triangle with vertices of three colors
         glClearBufferfv(GL_COLOR, 0, white);
        glPolygonMode(GL_FRONT, GL_FILL);
        glBindVertexArray(myVAO[colorTri]);
        glDrawArrays(GL_TRIANGLES, 0, 3);
        break;
    case 9:
        // Draw triangle with vertices of three colors (with different shader program)
        glClearBufferfv(GL_COLOR, 0, white);
        glUseProgram(shaderProgram3);
        glBindVertexArray(myVAO[colorTri]);
        glDrawArrays(GL_TRIANGLES, 0, 3);
        break;
    }

    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}

void my_setup_SceneData() {
    mySetupGeometries();

    GlShaderMgr::LoadShaderSource("Chapter1Figs.glsl");

    // Either of the next two options will work for Chatper1Figs
#if 1
    shaderProgram1 = GlShaderMgr::CompileAndLinkProgram("vertexShader_Chap1Figs", "fragmentShader_Chap1Figs");
    shaderProgram2 = GlShaderMgr::CompileAndLinkProgram("vertexShader_FlatChap1Figs", "fragmentShader_FlatChap1Figs");
    shaderProgram3 = GlShaderMgr::CompileAndLinkProgram("vertexShader_Chap1Figs", "fragmentShader_BrightColor");
#else
    unsigned int vertexShader = GlShaderMgr::CompileShader("vertexShader_Chap1Figs");
    unsigned int fragmentShader = GlShaderMgr::CompileShader("fragmentShader_Chap1Figs");
    unsigned int vertexShaderFlat = GlShaderMgr::CompileShader("vertexShader_FlatChap1Figs");
    unsigned int fragmentShaderFlat = GlShaderMgr::CompileShader("fragmentShader_FlatChap1Figs");
    unsigned int fragmentShaderBright = GlShaderMgr::CompileShader("fragmentShader_BrightColor");
    unsigned int shaders1[2] = { vertexShader, fragmentShader };
    shaderProgram1 = GlShaderMgr::LinkShaderProgram(2, shaders1);
    unsigned int shaders2[2] = { vertexShaderFlat, fragmentShaderFlat };
    shaderProgram2 = GlShaderMgr::LinkShaderProgram(2, shaders2);
    unsigned int shaders3[2] = { vertexShader, fragmentShaderBright };
    shaderProgram3 = GlShaderMgr::LinkShaderProgram(2, shaders3);
#endif

    // Free up shader compilation resources that are no longer needed.
    GlShaderMgr::FinalizeCompileAndLink();
     
    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}

// *******************************************************
// Process all key press events.
// This routine is called each time a key is pressed or released.
// *******************************************************
void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods) {
	if (action == GLFW_RELEASE) {
		return;			// Ignore key up (key release) events
	}
	if (key == GLFW_KEY_ESCAPE || key == GLFW_KEY_X) {
		glfwSetWindowShouldClose(window, true);
	}
	else if (key == GLFW_KEY_SPACE) {
		CurrentMode = (CurrentMode+1) % 10;	// Takes on values from 0 to 9
	}
}


// *************************************************
// This function is called with the graphics window is first created,
//    and again whenever it is resized.
// The Projection View Matrix is typically set here.
//    But this program does not use any transformations or matrices.
// *************************************************
void window_size_callback(GLFWwindow* window, int width, int height) {
	glViewport(0, 0, width, height);		// Draw into entire window
}

void my_setup_OpenGL() {
	
	glEnable(GL_DEPTH_TEST);	// Enable depth buffering
	glDepthFunc(GL_LEQUAL);		// Useful for multipass shaders

	// Set polygon drawing mode for front and back of each triangle
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	// glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	// glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);

	// Disable backface culling to render both sides of triangles
	// glDisable(GL_CULL_FACE);

// TRY IT OUT: How do the results look different if you disable the next block
// of code. (By changing "#if 1" to "#if 0"
#if 1
	// The following commands should induce OpenGL to create round points and 
	//	antialias points and lines.  (This is implementation dependent unfortunately.)
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);	// Make round points, not square points
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);		// Antialias the lines
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
#endif

	// Specify the diameter of points, and the width of lines. (Implementation dependent.)
	//   Permissible values are 1 and greater.
	// TRY IT OUT: Experiment with increasing and decreasing these values.
	glPointSize(8);
	glLineWidth(5);
	
}

void error_callback(int error, const char* description)
{
	// Print error
	fputs(description, stderr);
}

void setup_callbacks(GLFWwindow* window) {
	// Set callback function for resizing the window
	glfwSetFramebufferSizeCallback(window, window_size_callback);

	// Set callback for key up/down/repeat events
	glfwSetKeyCallback(window, key_callback);

	// Set callbacks for mouse movement (cursor position) and mouse botton up/down events.
	// glfwSetCursorPosCallback(window, cursor_pos_callback);
	// glfwSetMouseButtonCallback(window, mouse_button_callback);
}

// **********************
// Here is the main program
// **********************

int main() {
	glfwSetErrorCallback(error_callback);	// Supposed to be called in event of errors. (doesn't work?)
	glfwInit();
	//glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	//glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	//glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
	//glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);

	const int initWidth = 600;
	const int initHeight = 900;
    glfwWindowHint(GLFW_SAMPLES, 9); // Invoke MSAA
    glEnable(GL_MULTISAMPLE);
	GLFWwindow* window = glfwCreateWindow(initWidth, initHeight, "Chapter1Figs", NULL, NULL);
	if (window == NULL) {
		fprintf(stdout, "Failed to create GLFW window!\n");
		glfwTerminate();
		return -1;
	}
	glfwMakeContextCurrent(window);

	glewInit();

	// Print info of GPU and supported OpenGL version
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("OpenGL version supported %s\n", glGetString(GL_VERSION));
#ifdef GL_SHADING_LANGUAGE_VERSION
	printf("Supported GLSL version is %s.\n", (char *)glGetString(GL_SHADING_LANGUAGE_VERSION));
#endif
    printf("Using GLEW version %s.\n", glewGetString(GLEW_VERSION));

	printf("------------------------------\n");
	printf("Press space bar to cycle thru images.\n");
	printf("Press ESCAPE or 'X' or 'x' to exit.\n");
	
    setup_callbacks(window);
    window_size_callback(window, initWidth, initHeight);

	// Initialize OpenGL, the scene and the shaders
    my_setup_OpenGL();
	my_setup_SceneData();
 
    // Loop while program is not terminated.
	while (!glfwWindowShouldClose(window)) {
	
		myRenderScene();				// Render into the current buffer
		glfwSwapBuffers(window);		// Displays what was just rendered (using double buffering).

		// Poll events (key presses, mouse events)
		glfwWaitEvents();					// Use this if no animation.
		//glfwWaitEventsTimeout(1.0/60.0);	// Use this to animate at 60 frames/sec (timing is NOT reliable)
		// glfwPollEvents();				// Use this version when animating as fast as possible
	}

	glfwTerminate();
	return 0;
}

// If an error is found, it could have been caused by any command since the
//   previous call to check_for_opengl_errors()
// To find what generated the error, you can try adding more calls to
//   check_for_opengl_errors().
char errNames[8][36] = {
    "Unknown OpenGL error",
    "GL_INVALID_ENUM", "GL_INVALID_VALUE", "GL_INVALID_OPERATION",
    "GL_INVALID_FRAMEBUFFER_OPERATION", "GL_OUT_OF_MEMORY",
    "GL_STACK_UNDERFLOW", "GL_STACK_OVERFLOW" };
bool check_for_opengl_errors() {
    int numErrors = 0;
    GLenum err;
    while ((err = glGetError()) != GL_NO_ERROR) {
        numErrors++;
        int errNum = 0;
        switch (err) {
        case GL_INVALID_ENUM:
            errNum = 1;
            break;
        case GL_INVALID_VALUE:
            errNum = 2;
            break;
        case GL_INVALID_OPERATION:
            errNum = 3;
            break;
        case GL_INVALID_FRAMEBUFFER_OPERATION:
            errNum = 4;
            break;
        case GL_OUT_OF_MEMORY:
            errNum = 5;
            break;
        case GL_STACK_UNDERFLOW:
            errNum = 6;
            break;
        case GL_STACK_OVERFLOW:
            errNum = 7;
            break;
        }
        printf("OpenGL ERROR: %s.\n", errNames[errNum]);
    }
    return (numErrors != 0);
}